#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>

#ifndef PORT
#define PORT 10000
#endif

#define MALLOC_SIZE 128

char* readpw()
{
	char *pw = malloc(MALLOC_SIZE);
	FILE *f;
	int n;

	f = fopen("passwd", "r");
	if (f == NULL) {
		printf("internal error: cannot open password file\n");
		goto error;
	}
	n = fread(pw, 1, MALLOC_SIZE, f);

	if (n < MALLOC_SIZE && ferror(f)) {
		printf("internal error: cannot read pasword file\n");
		goto error;
	}
	if (n < 17) {
		printf("internal error: malformed password file\n");
		goto error;
	}
	pw[n - 1] = '\0';

	return pw;

error:
	free(pw);
	return NULL;
}

void child()
{
	char *cmd, *scan;
	int n;

	for (;;) {
		cmd = malloc(MALLOC_SIZE);
		n = read(0, cmd, MALLOC_SIZE);
		if (n < 0) {
			printf("read error\n");
			return;
		}
		if (n == 0) {
			printf("goodbye\n");
			return;
		}
		if (n < 4) {
			printf("format error\n");
			return;
		}
		scan = index(cmd, '\n');
		if (scan != NULL)
			*scan = '\0';
		if (!strncmp("PWD ", cmd, 4)) {
			char *pw = readpw();
			if (pw == NULL)
				return;
			// the first 16 bytes contain the username
			if (!strcmp(pw + 16, cmd + 4)) {
				printf("Password OK!\n");
			} else {
				printf("Invalid password!\n");
			}
			free(cmd);
			free(pw);
		} else if (!strncmp("QUIT", cmd, 4)) {
			printf("goodbye\n");
			free(cmd);
			return;
		} else {
			printf("Unrecognized command: %s\n", cmd);
			free(cmd);
		}
	}

}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, SIG_IGN);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			dup(con);
			close(1);
			dup(con);
			close(2);
			dup(con);
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;
}
